#pragma once
#include "../include/rational.hpp"
#include "../include/operator.hpp"
#include "../include/binsaver.hpp"
#include "../include/textsaver.hpp"

#include "../include/addition.hpp"
#include "../include/subtraction.hpp"
#include "../include/multiplication.hpp"
#include "../include/division.hpp"
#include "../include/modulo.hpp"
#include "../include/exponencial.hpp"
#include "../include/logarithm.hpp"

#include <map>
#include <cmath>
#include <algorithm>
#include <sstream>
#include <iterator>
#include <cctype>
//https://www.wikiwand.com/en/Lexical_analysis
//https://courses.fit.cvut.cz/BI-AAG/lectures/bi-aag-07-bezkontextove_gramatiky.pdf

/**
 * @class CParser
 * @brief Singleton class responsible for converting input into results.
 */
class CParser {
public:
    /**
     * @brief Default constructor for CParser class.
     */
    CParser ();

    /**
     * @brief Destructor for CParser class. Clears the parser data.
     */
    ~CParser();
    void reset();

    /**
     * @brief Clear the parser data.
     */
    void clear();
    void convert( const std::string &s );

    /**
     * @brief Divide a decimal string by 2 and return the next bit for binary conversion.
     *
     * @param s The decimal string to be divided.
     * @return The next bit for binary conversion.
     */
    static bool stringDivideDecimal( std::string &s );

    bool m_continue;
    std::string m_output;
private:
    enum TOKEN { //by priority
        START,
        END,
        NUMBER,
        LABEL,
        COMMAND,     //.!
        DEREFERENCE, // $
        OPEN,        // (
        CLOSE,       // )
        LN,          // &
        EXP,         // ^
        DIVIDE,      // /
        MULTIPLY,    //.*
        MOD,         // %
        ADD,         // +
        SUBTRACT,    // -
        SAVE,        // =
        DOT,
        WHITESPACE,
        SYNTAX_ERROR
    };

    enum CMD {
        QUIT_CMD,
        LOAD_CMD,
        SAVE_CMD,
        PRECISSION_CMD,
        HELP_CMD,
        ROUND_CMD,
        UNKNOWN_CMD,
        NONE_CMD
    };

    CParser::CMD recognizeCommand( uint64_t commandPos ) const;

    CParser::CMD parseCommand( uint64_t &startPos ) const;

    bool executeCommand( CParser::CMD cmd );

    /**
     * @brief Tokenize the input string.
     *
     * This function analyzes the input string and breaks it down into lexemes,
     * classifying each lexeme into different types such as whitespace, number, label,
     * command, or operators.
     *
     * @return True if tokenization is successful, false otherwise.
     */
    bool tokenize();
    
    /**
     * @brief Perform grammar preprocessing on the raw lexemes.
     *
     * This function strips whitespaces from the raw lexemes and processes numbers
     * by creating rational numbers and handling various cases such as decimals,
     * signs, and dereferencing variables.
     *
     * @return True if grammar preprocessing is successful, false otherwise.
     */
    bool grammarPreprocess(); // add bracket based on priority

    /**
     * @brief Parses the input expression and constructs a syntax tree.
     *
     * This function recursively parses the input expression and constructs a syntax tree
     * based on the processed lexemes. It handles invalid semantics and builds the
     * tree of operators and operands.
     *
     * @param readIndex Index of opening bracket.
     * @return A pointer to the root of the syntax tree.
     */
    CPrintable *parse( uint64_t &readPos );

    /**
     * @brief Get the priority of an operator.
     * 
     * @param t The operator token.
     * @return The priority of the operator.
     */
    uint64_t getOperatorPriority( TOKEN t ) const;

    /**
     * @brief Show the help command and return true.
     * 
     * @return True.
     */
    bool helpCommand();
    
    /**
     * @brief Reads a token from the input string.
     *
     * This function reads a token from the input string based on the provided index. It starts
     * from the position indicated by the index and continues until it encounters a space character
     * or reaches the end of the token.
     *
     * @param i The index of the token in the raw lexemes.
     * @return The extracted token as a string.
     */
    std::string readToken( uint64_t i ) const; // position in raw tokens

    /**
     * @brief Converts a CRational number to a string representation.
     *
     * This function converts a CRational number to its string representation. The string
     * representation includes the numerator and, if the denominator is not equal to 1, the
     * fraction part in the format "numerator / denominator".
     *
     * @param r A pointer to the CRational number to convert.
     * @return The string representation of the CRational number.
     */
    std::string numberToString( CRational *r ) const;

    /**
     * @brief Parse a number from the input string and convert it to a CRational object.
     *
     * @param i The index of the token in the raw lexems vector.
     * @param isDecimal Flag indicating if the number is a decimal.
     * @param hasSign Flag indicating if the number has a sign.
     * @return A pointer to the parsed CRational object.
     */
    CRational *parseNumber( uint64_t i, bool isDecimal, bool hasSign ) const;

    /**
     * @brief Creates a new instance of an operator based on the given token.
     *
     * This function creates a new instance of a COperator-based class based on the provided token.
     * The specific type of the operator is determined by the token value.
     *
     * @param t The token representing the operator.
     * @return A pointer to the newly created operator instance, or nullptr if the token is not a valid operator.
     */
    COperator *spawnOperator( TOKEN t ) const;

    /**
     * @brief Recognize the type of a character token.
     * 
     * @param c The character token.
     * @return The type of the token.
     */
    TOKEN recognize( char c ) const;

    /**
     * @brief Container for recognised lexems and their position in input.
     */
    std::vector<std::pair<CParser::TOKEN, int64_t>> m_rawLexems;

    /**
     * @brief Container for relevant lexems and parsed numbers
     */
    std::vector<std::pair<CParser::TOKEN, CRational *>> m_processedLexems;

    /**
     * @brief Copy of input string
     */
    std::string m_input;
    std::string m_saveInto;
    std::unique_ptr<CPrintable> m_head;
    std::map<std::string, CRational> m_variables;
    std::unique_ptr<CSaver> m_saver;
};